"""
Module defining the main matplotlib Artist for network/tree edges, EdgeCollection.

Some supporting functions are also defined here.
"""

from typing import (
    Sequence,
    Optional,
    Any,
)
from math import atan2, cos, pi, sin
from collections import defaultdict
import numpy as np
import pandas as pd
import matplotlib as mpl

from ..typing import (
    Pair,
    LeafProperty,
)
from ..utils.matplotlib import (
    _compute_mid_coord_and_rot,
    _stale_wrapper,
    _forwarder,
)
from ..style import (
    rotate_style,
)
from ..label import LabelCollection
from ..vertex import VertexCollection
from .arrow import EdgeArrowCollection
from .geometry import (
    _compute_loops_per_angle,
    _fix_parallel_edges_straight,
    _compute_loop_path,
    _compute_edge_path,
)


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class EdgeCollection(mpl.collections.PatchCollection):
    """Artist for a collection of edges within a network/tree.

    This artist is derived from PatchCollection with a few notable differences:
      - It udpdates ends of each edge based on the vertex borders.
      - It may contain edge labels as a child (a LabelCollection).
      - For directed graphs, it contains arrows as a child (an EdgeArrowCollection).

    This class is not designed to be instantiated directly but rather by internal
    iplotx functions such as iplotx.network. However, some of its methods can be
    called directly to edit edge style after the initial draw.
    """

    def __init__(
        self,
        patches: Sequence[mpl.patches.Patch],
        vertex_ids: Sequence[tuple],
        vertex_collection: VertexCollection,
        *args,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        arrow_transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        directed: bool = False,
        style: Optional[dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        """Initialise an EdgeCollection.

        Parameters:
            patches: A sequence (usually, list) of matplotlib `Patch`es describing the edges.
            vertex_ids: A sequence of pairs `(v1, v2)`, each defining the ids of vertices at the
                end of an edge.
            vertex_collection: The VertexCollection instance containing the Artist for the
                vertices. This is needed to compute vertex borders and adjust edges accordingly.
            transform: The matplotlib transform for the edges, usually transData.
            arrow_transform: The matplotlib transform for the arrow patches. This is not the
                *offset_transform* of arrows, which is set equal to the edge transform (previous
                parameter). Instead, it specifies how arrow size scales, similar to vertex size.
                This is usually the identity transform.
            directed: Whether the graph is directed (in which case arrows are drawn, possibly
                with zero size or opacity to obtain an "arrowless" effect).
            style: The edge style (subdictionary: "edge") to use at creation.
        """
        kwargs["match_original"] = True
        self._vertex_ids = vertex_ids

        self._vertex_collection = vertex_collection
        self._style = style if style is not None else {}
        self._labels = kwargs.pop("labels", None)
        self._directed = directed
        self._arrow_transform = arrow_transform
        if "cmap" in self._style:
            kwargs["cmap"] = self._style["cmap"]
            kwargs["norm"] = self._style["norm"]

        # NOTE: This should also set the transform
        super().__init__(patches, transform=transform, *args, **kwargs)

        # This is important because it prepares the right flags for scalarmappable
        self.set_facecolor("none")

        if self.directed:
            self._arrows = EdgeArrowCollection(
                self,
                transform=self._arrow_transform,
            )
        if self._labels is not None:
            style = self._style.get("label", {})
            self._label_collection = LabelCollection(
                self._labels,
                style=style,
                transform=transform,
            )

    def get_children(self) -> tuple:
        children = []
        if hasattr(self, "_arrows"):
            children.append(self._arrows)
        if hasattr(self, "_label_collection"):
            children.append(self._label_collection)
        return tuple(children)

    def set_figure(self, fig) -> None:
        super().set_figure(fig)
        self._update_paths()
        # NOTE: This sets the correct offsets in the arrows,
        # but not the correct sizes (see below)
        self._update_children()
        for child in self.get_children():
            # NOTE: This sets the sizes with correct dpi scaling in the arrows
            child.set_figure(fig)

    def _update_children(self):
        self._update_arrows()
        self._update_labels()

    @property
    def directed(self) -> bool:
        """Whether the network is directed."""
        return self._directed

    @directed.setter
    def directed(self, value) -> None:
        """Setter for the directed property.

        Changing this property triggers the addition/removal of arrows from the plot.
        """
        value = bool(value)
        if self._directed != value:
            # Moving to undirected, remove arrows
            if not value:
                self._arrows.remove()
                del self._arrows
            # Moving to directed, create arrows
            else:
                self._arrows = EdgeArrowCollection(
                    self,
                    transform=self._arrow_transform,
                )

            self._directed = value
            # NOTE: setting stale to True should trigger a redraw as soon as needed
            # and that will update children. We might need to verify that.
            self.stale = True

    def set_array(self, A) -> None:
        """Set the array for cmap/norm coloring."""
        # Preserve the alpha channel
        super().set_array(A)
        # Alpha needs to be kept separately
        if self.get_alpha() is None:
            self.set_alpha(self.get_edgecolor()[:, 3])
        # This is necessary to ensure edgecolors are bool-flagged correctly
        self.set_edgecolor(None)

    def update_scalarmappable(self) -> None:
        """Update colors from the scalar mappable array, if any.

        Assign edge colors from a numerical array, and match arrow colors
        if the graph is directed.
        """
        # NOTE: The superclass also sets stale = True
        super().update_scalarmappable()
        # Now self._edgecolors has the correct colorspace values
        if hasattr(self, "_arrows"):
            self._arrows.set_colors(self.get_edgecolor())

    def get_labels(self) -> Optional[LabelCollection]:
        """Get LabelCollection artist for labels if present."""
        if hasattr(self, "_label_collection"):
            return self._label_collection
        return None

    def get_mappable(self):
        """Return mappable for colorbar."""
        return self

    def _get_adjacent_vertices_info(self):
        index = self._vertex_collection.get_index()
        index = pd.Series(
            np.arange(len(index)),
            index=index,
        )

        voffsets = []
        vpaths = []
        vsizes = []
        for v1, v2 in self._vertex_ids:
            # NOTE: these are in the original layout coordinate system
            # not cartesianised yet.
            offset1 = self._vertex_collection.get_layout().values[index[v1]]
            offset2 = self._vertex_collection.get_layout().values[index[v2]]
            voffsets.append((offset1, offset2))

            path1 = self._vertex_collection.get_paths()[index[v1]]
            path2 = self._vertex_collection.get_paths()[index[v2]]
            vpaths.append((path1, path2))

            # NOTE: This needs to be computed here because the
            # VertexCollection._transforms are reset each draw in order to
            # accomodate for DPI changes on the canvas
            size1 = self._vertex_collection.get_sizes_dpi()[index[v1]]
            size2 = self._vertex_collection.get_sizes_dpi()[index[v2]]
            vsizes.append((size1, size2))

        return {
            "ids": self._vertex_ids,
            "offsets": voffsets,
            "paths": vpaths,
            "sizes": vsizes,
        }

    def _update_paths(self, transform=None):
        """Compute paths for the edges.

        Loops split the largest wedge left open by other
        edges of that vertex. The algo is:
        (i) Find what vertices each loop belongs to
        (ii) While going through the edges, record the angles
             for vertices with loops
        (iii) Plot each loop based on the recorded angles
        """
        vinfo = self._get_adjacent_vertices_info()
        vids = vinfo["ids"]
        vcenters = vinfo["offsets"]
        vpaths = vinfo["paths"]
        vsizes = vinfo["sizes"]
        loopmaxangle = pi / 180.0 * self._style.get("loopmaxangle", 60.0)

        if transform is None:
            transform = self.get_transform()
        trans = transform.transform
        trans_inv = transform.inverted().transform

        # 1. Make a list of vertices with loops, and store them for later
        loop_vertex_dict = defaultdict(lambda: dict(indices=[], edge_angles=[]))
        for i, (v1, v2) in enumerate(vids):
            # Postpone loops (step 3)
            if v1 == v2:
                loop_vertex_dict[v1]["indices"].append(i)

        # 2. Make paths for non-loop edges
        # NOTE: keep track of parallel edges to offset them
        parallel_edges = defaultdict(list)
        paths = []
        for i, (v1, v2) in enumerate(vids):
            # Postpone loops (step 3)
            if v1 == v2:
                paths.append(None)
                continue

            # Coordinates of the adjacent vertices, in data coords
            vcoord_data = vcenters[i]

            # Vertex paths in figure (default) coords
            vpath_fig = vpaths[i]

            # Vertex size
            vsize_fig = vsizes[i]

            # Leaf rotation
            edge_stylei = rotate_style(self._style, index=i, key=(v1, v2))
            if edge_stylei.get("curved", False):
                tension = edge_stylei.get("tension", 5)
                ports = edge_stylei.get("ports", (None, None))
            else:
                tension = 0
                ports = None

            waypoints = edge_stylei.get("waypoints", "none")
            if waypoints != "none":
                ports = edge_stylei.get("ports", (None, None))

            # Compute actual edge path
            path, angles = _compute_edge_path(
                vcoord_data,
                vpath_fig,
                vsize_fig,
                trans,
                trans_inv,
                tension=tension,
                waypoints=waypoints,
                ports=ports,
                layout_coordinate_system=self._vertex_collection.get_layout_coordinate_system(),
            )

            offset = edge_stylei.get("offset", 0)
            if np.isscalar(offset):
                if offset == 0:
                    offset = (0, 0)
                else:
                    vd_fig = trans(vcoord_data[1]) - trans(vcoord_data[0])
                    vd_fig /= np.linalg.norm(vd_fig)
                    vrot = vd_fig @ np.array([[0, -1], [1, 0]])
                    offset = offset * vrot
            offset = np.asarray(offset, dtype=float)
            # Scale by dpi
            dpi = self.figure.dpi if hasattr(self, "figure") else 72.0
            offset *= dpi / 72.0
            if (offset != 0).any():
                path.vertices[:] = trans_inv(trans(path.vertices) + offset)

            # Collect angles for this vertex, to be used for loops plotting below
            if v1 in loop_vertex_dict:
                loop_vertex_dict[v1]["edge_angles"].append(angles[0])
            if v2 in loop_vertex_dict:
                loop_vertex_dict[v2]["edge_angles"].append(angles[1])

            # Add the path for this non-loop edge
            paths.append(path)
            # FIXME: curved parallel edges depend on the direction of curvature...!
            parallel_edges[(v1, v2)].append(i)

        # Fix parallel edges
        # If none found, empty the dictionary already
        if (len(parallel_edges) == 0) or (max(parallel_edges.values(), key=len) == 1):
            parallel_edges = {}
        if not self._style.get("curved", False):
            while len(parallel_edges) > 0:
                (v1, v2), indices = parallel_edges.popitem()
                indices_inv = parallel_edges.pop((v2, v1), [])
                ntot = len(indices) + len(indices_inv)
                if ntot > 1:
                    _fix_parallel_edges_straight(
                        paths,
                        indices,
                        indices_inv,
                        trans,
                        trans_inv,
                        paralleloffset=self._style.get("paralleloffset", 3),
                    )

        # 3. Deal with loops at the end
        for vid, ldict in loop_vertex_dict.items():
            vpath = vpaths[ldict["indices"][0]][0]
            vsize = vsizes[ldict["indices"][0]][0]
            vcoord_fig = trans(vcenters[ldict["indices"][0]][0])
            nloops = len(ldict["indices"])
            edge_angles = ldict["edge_angles"]

            # The space between the existing angles is where we can fit the loops
            # One loop we can fit in the largest wedge, multiple loops we need
            nloops_per_angle = _compute_loops_per_angle(nloops, edge_angles)

            idx = 0
            for theta1, theta2, nloops in nloops_per_angle:
                # Angular size of each loop in this wedge
                delta = (theta2 - theta1) / nloops

                # Iterate over individual loops
                for j in range(nloops):
                    thetaj1 = theta1 + j * delta + max(delta - loopmaxangle, 0) / 2
                    thetaj2 = thetaj1 + min(delta, loopmaxangle)

                    # Get the path for this loop
                    path = _compute_loop_path(
                        vcoord_fig,
                        vpath,
                        vsize,
                        thetaj1,
                        thetaj2,
                        trans_inv,
                        looptension=self._style.get("looptension", 2.5),
                    )
                    paths[ldict["indices"][idx]] = path
                    idx += 1

        self._paths = paths

    def _update_labels(self):
        if self._labels is None:
            return

        style = self._style.get("label", None) if self._style is not None else {}
        transform = self.get_transform()
        trans = transform.transform

        offsets = []
        if not style.get("rotate", True):
            rotations = []
        for path in self._paths:
            offset, rotation = _compute_mid_coord_and_rot(path, trans)
            offsets.append(offset)
            if not style.get("rotate", True):
                rotations.append(rotation)

        self._label_collection.set_offsets(offsets)
        if not style.get("rotate", True):
            self._label_collection.set_rotations(rotations)

    def _update_arrows(
        self,
    ) -> None:
        """Extract the start and/or end angles of the paths to compute arrows.

        Parameters:
            which: Which end of the edge to put an arrow on. Currently only "end" is accepted.

        NOTE: This function does *not* update the arrow sizes/_transforms to the correct dpi
        scaling. That's ok since the correct dpi scaling is set whenever there is a different
        figure (before first draw) and whenever a draw is called.
        """
        if not hasattr(self, "_arrows"):
            return

        transform = self.get_transform()
        trans = transform.transform

        for i, epath in enumerate(self.get_paths()):
            # Offset the arrow to point to the end of the edge
            self._arrows._offsets[i] = epath.vertices[-1]

            # Rotate the arrow to point in the direction of the edge
            apath = self._arrows._paths[i]
            # NOTE: because the tip of the arrow is at (0, 0) in patch space,
            # in theory it will rotate around that point already
            v2 = trans(epath.vertices[-1])
            v1 = trans(epath.vertices[-2])
            dv = v2 - v1
            theta = atan2(*(dv[::-1]))
            theta_old = self._arrows._angles[i]
            dtheta = theta - theta_old
            mrot = np.array([[cos(dtheta), sin(dtheta)], [-sin(dtheta), cos(dtheta)]])
            apath.vertices = apath.vertices @ mrot
            self._arrows._angles[i] = theta

    @_stale_wrapper
    def draw(self, renderer):
        # Visibility affects the children too
        if not self.get_visible():
            return

        self._update_paths()
        # This sets the arrow offsets
        self._update_children()

        super().draw(renderer)
        for child in self.get_children():
            # This sets the arrow sizes with dpi scaling
            child.draw(renderer)

    def get_ports(self) -> Optional[LeafProperty[Pair[Optional[str]]]]:
        """Get the ports for all edges.

        Returns:
            The ports for the edges, as a pair of strings or None for each edge. If None, it
            means all edges are free.
        """
        return self._style.get("ports", None)

    def set_ports(self, ports: Optional[LeafProperty[Pair[Optional[str]]]]) -> None:
        """Set new ports for the edges.

        Parameters:
            ports: A pair of ports strings for each edge. Each port can be None to mean free
                edge end.
        """
        if ports is None:
            del self._style["ports"]
        else:
            self._style["ports"] = ports
        self.stale = True

    def get_tension(self) -> Optional[LeafProperty[float]]:
        """Get the tension for the edges.

        Returns:
            The tension for the edges. If None, the edges are straight.
        """
        return self._style.get("tension", None)

    def set_tension(self, tension: Optional[LeafProperty[float]]) -> None:
        """Set new tension for the edges.

        Parameters:
            tension: The tension to use for curved edges. If None, the edges become straight.

        Note: This function does not set self.set_curved(True) automatically. If you are
        unsure whether that property is set already, you should call both functions.

        Example:
            # Set curved edges with different tensions
            >>> network.get_edges().set_curved(True)
            >>> network.get_edges().set_tension([1, 0.5])

            # Set straight edges
            # (the latter call is optional but helps readability)
            >>> network.get_edges().set_curved(False)
            >>> network.get_edges().set_tension(None)

        """
        if tension is None:
            del self._style["tension"]
        else:
            self._style["tension"] = tension
        self.stale = True

    get_tensions = get_tension
    set_tensions = set_tension

    def get_curved(self) -> bool:
        """Get whether the edges are curved or not.

        Returns:
            A bool that is True if the edges are curved, False if they are straight.
        """
        return self._style.get("curved", False)

    def set_curved(self, curved: bool) -> None:
        """Set whether the edges are curved or not.

        Parameters:
            curved: Whether the edges should be curved (True) or straight (False).

        Note: If you want only some edges to be curved, set curved to True and set tensions to
        0 for the straight edges.
        """
        self._style["curved"] = bool(curved)
        self.stale = True

    def get_loopmaxangle(self) -> Optional[float]:
        """Get the maximum angle for loops.

        Returns:
            The maximum angle in degrees that a loop can take. If None, the default is 60.
        """
        return self._style.get("loopmaxangle", 60)

    def set_loopmaxangle(self, loopmaxangle: float) -> None:
        """Set the maximum angle for loops.

        Parameters:
            loopmaxangle: The maximum angle in degrees that a loop can take.
        """
        self._style["loopmaxangle"] = loopmaxangle
        self.stale = True

    def get_looptension(self) -> Optional[float]:
        """Get the tension for loops.

        Returns:
            The tension for loops. If None, the default is 2.5.
        """
        return self._style.get("looptension", 2.5)

    def set_looptension(self, looptension: Optional[float]) -> None:
        """Set new tension for loops.

        Parameters:
            looptension: The tension to use for loops. If None, the default is 2.5.
        """
        if looptension is None:
            del self._style["looptension"]
        else:
            self._style["looptension"] = looptension
        self.stale = True

    def get_offset(self) -> Optional[float]:
        """Get the offset for parallel straight edges.

        Returns:
            The offset in points for parallel straight edges. If None, the default is 3.
        """
        return self._style.get("offset", 3)

    def set_offset(self, offset: Optional[float]) -> None:
        """Set the offset for parallel straight edges.

        Parameters:
            offset: The offset in points for parallel straight edges. If None, the default is 3.
        """
        if offset is None:
            del self._style["offset"]
        else:
            self._style["offset"] = offset
        self.stale = True


def make_stub_patch(**kwargs):
    """Make a stub undirected edge patch, without actual path information."""
    kwargs["clip_on"] = kwargs.get("clip_on", True)
    if ("color" in kwargs) and ("edgecolor" not in kwargs):
        kwargs["edgecolor"] = kwargs.pop("color")

    # Edges are always hollow, because they are not closed paths
    # NOTE: This is supposed to cascade onto what boolean flags are set
    # for color mapping (Colorizer)
    kwargs["facecolor"] = "none"

    # Forget specific properties that are not supported here
    forbidden_props = [
        "arrow",
        "label",
        "curved",
        "tension",
        "waypoints",
        "ports",
        "looptension",
        "loopmaxangle",
        "offset",
        "paralleloffset",
        "cmap",
    ]
    for prop in forbidden_props:
        if prop in kwargs:
            kwargs.pop(prop)

    # NOTE: the path is overwritten later anyway, so no reason to spend any time here
    art = mpl.patches.PathPatch(
        mpl.path.Path([[0, 0]]),
        **kwargs,
    )
    return art



"""
Multipartite layout
===================

This example from networkx shows how to use simple vertex styling.
"""

import itertools
import matplotlib.pyplot as plt
import networkx as nx
import iplotx as ipx

subset_sizes = [5, 5, 4, 3, 2, 4, 4, 3]
subset_color = [
    "gold",
    "violet",
    "violet",
    "violet",
    "violet",
    "limegreen",
    "limegreen",
    "darkorange",
]


def multilayered_graph(*subset_sizes):
    extents = nx.utils.pairwise(itertools.accumulate((0,) + subset_sizes))
    layers = [range(start, end) for start, end in extents]
    G = nx.Graph()
    for i, layer in enumerate(layers):
        G.add_nodes_from(layer, layer=i)
    for layer1, layer2 in nx.utils.pairwise(layers):
        G.add_edges_from(itertools.product(layer1, layer2))
    return G


G = multilayered_graph(*subset_sizes)
color = [subset_color[data["layer"]] for v, data in G.nodes(data=True)]
pos = nx.multipartite_layout(G, subset_key="layer")

fig, ax = plt.subplots(figsize=(6, 4))
ipx.plot(
    G,
    layout=pos,
    style={
        "vertex": {
            "facecolor": color,
        },
    },
    ax=ax,
)



"""
Plot a house
============

This example from networkx shows how to integrate network attributes/data and plotting.
"""

import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

G = nx.house_graph()
fig, ax = plt.subplots()
nx.set_node_attributes(
    G, {0: (0, 0), 1: (1, 0), 2: (0, 1), 3: (1, 1), 4: (0.5, 2.0)}, "pos"
)
nx.set_node_attributes(
    G,
    {
        n: {
            "size": 60 if n != 4 else 40,
            "color": "tab:blue" if n != 4 else "tab:orange",
        }
        for n in G.nodes()
    },
)
ipx.plot(
    G,
    ax=ax,
    layout="pos",
    style={
        "edge": {
            "alpha": 0.5,
            "linewidth": 6,
        },
        "vertex": {
            "size": G.nodes.data("size"),
            "facecolor": G.nodes.data("color"),
            "edgecolor": "k",
        },
    },
)
ax.margins(0.17)
plt.tight_layout()



from io import StringIO
import os
import unittest
import pytest
import numpy as np
import matplotlib as mpl

try:
    import skbio
except ImportError:
    raise unittest.SkipTest("skbio not found, skipping tests")

mpl.use("agg")
import matplotlib.pyplot as plt

import iplotx as ipx

from utils import image_comparison


class TreeTestRunner(unittest.TestCase):

    @property
    def small_tree(self):
        from skbio import TreeNode

        tree = TreeNode.read(
            StringIO(
                "((),((),(((),()),((),()))));",
            )
        )

        return tree

    @image_comparison(baseline_images=["tree_basic"], remove_text=True)
    def test_basic(self):
        tree = self.small_tree
        fig, ax = plt.subplots(figsize=(3, 3))
        ipx.plotting.tree(
            tree=tree,
            ax=ax,
            layout="horizontal",
        )

    @image_comparison(baseline_images=["tree_radial"], remove_text=True)
    def test_radial(self):
        tree = self.small_tree
        fig, ax = plt.subplots(figsize=(3, 3))
        ipx.plotting.tree(
            tree=tree,
            ax=ax,
            layout="radial",
            aspect=1,
        )

    @image_comparison(baseline_images=["leaf_labels"], remove_text=True)
    def test_leaf_labels(self):
        tree = self.small_tree
        leaf_labels = {leaf: str(i + 1) for i, leaf in enumerate(tree.tips())}

        fig, ax = plt.subplots(figsize=(4, 4))
        ipx.plotting.tree(
            tree=tree,
            ax=ax,
            layout="horizontal",
            vertex_labels=leaf_labels,
            margins=0.1,
        )

    @image_comparison(baseline_images=["leaf_labels_hmargin"], remove_text=True)
    def test_leaf_labels_hmargin(self):
        tree = self.small_tree
        leaf_labels = {leaf: str(i + 1) for i, leaf in enumerate(tree.tips())}

        fig, ax = plt.subplots(figsize=(4, 4))
        ipx.plotting.tree(
            tree=tree,
            ax=ax,
            layout="horizontal",
            vertex_labels=leaf_labels,
            vertex_label_hmargin=[10, 22],
            margins=(0.15, 0),
        )



from math import atan2
import numpy as np


# See also this link for the general answer (using scipy to compute coefficients):
# https://stackoverflow.com/questions/12643079/b%C3%A9zier-curve-fitting-with-scipy
def _evaluate_squared_bezier(points, t):
    """Evaluate a squared Bezier curve at t."""
    p0, p1, p2 = points
    return (1 - t) ** 2 * p0 + 2 * (1 - t) * t * p1 + t**2 * p2


def _evaluate_cubic_bezier(points, t):
    """Evaluate a cubic Bezier curve at t."""
    p0, p1, p2, p3 = points
    return (
        (1 - t) ** 3 * p0
        + 3 * (1 - t) ** 2 * t * p1
        + 3 * (1 - t) * t**2 * p2
        + t**3 * p3
    )


def _evaluate_cubic_bezier_derivative(points, t):
    """Evaluate the derivative of a cubic Bezier curve at t."""
    p0, p1, p2, p3 = points
    # (dx / dt, dy / dt) is the parametric gradient
    # to get the angle from this, one can just atanh(dy/dt, dx/dt)
    # This is equivalent to computing the actual bezier curve
    # at low t, of course, which is the geometric interpretation
    # (obviously, division by t is irrelenant)
    return (
        3 * p0 * (1 - t) ** 2
        + 3 * p1 * (1 - t) * (-3 * t + 1)
        + 3 * p2 * t * (2 - 3 * t)
        + 3 * p3 * t**2
    )


def convex_hull(points):
    """Compute the convex hull of a set of 2D points.

    This is guaranteed to return the vertices clockwise.

    (Therefore, (v[i+1] - v[i]) rotated *left* by pi/2 points *outwards* of the convex hull.)
    """
    points = np.asarray(points)
    if len(points) < 3:
        return np.arange(len(points))

    hull_idx = None

    # igraph's should be faster in 2D
    try:
        import igraph

        hull_idx = igraph.convex_hull(list(points))
    except ImportError:
        pass

    # Otherwise, try scipy
    try:
        from scipy.spatial import ConvexHull

        # NOTE: scipy guarantees counterclockwise ordering in 2D
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.ConvexHull.html
        hull_idx = ConvexHull(points).vertices[::-1]
    except ImportError:
        pass

    # Last resort: our own Graham scan
    if hull_idx is None:
        hull_idx = _convex_hull_Graham_scan(points)

    return hull_idx


# see also: https://github.com/igraph/igraph/blob/075be76c92b99ca4c95ad9207bcc1af6d471c85e/src/misc/other.c#L116
# Compared to that C implementation, this is a bit more vectorised and messes less with memory as usual when
# optimising Python/numpy code
def _convex_hull_Graham_scan(points):
    """Compute the indices for the convex hull of a set of 2D points using Graham's scan algorithm.

    NOTE: This works from 3 points upwards, guaranteed clockwise.
    """
    points = np.asarray(points)

    # Find pivot (bottom left corner)
    miny_idx = np.flatnonzero(points[:, 1] == points[:, 1].min())
    pivot_idx = miny_idx[points[miny_idx, 0].argmin()]

    # Compute angles against that pivot, ensuring the pivot itself last
    angles = np.arctan2(
        points[:, 1] - points[pivot_idx, 1], points[:, 0] - points[pivot_idx, 0]
    )
    angles[pivot_idx] = np.inf

    # Sort points by angle
    order = np.argsort(angles)

    # Whenever two points have the same angle, keep the furthest one from the pivot
    # whenever an index is discarded from "order", set it to -1
    j = 0
    last_idx = order[0]
    pivot_idx = order[-1]
    for i in range(1, len(order)):
        next_idx = order[i]
        if angles[last_idx] == angles[next_idx]:
            dlast = np.linalg.norm(points[last_idx] - points[pivot_idx])
            dnext = np.linalg.norm(points[next_idx] - points[pivot_idx])
            # Ignore the new point, it's inside
            if dlast > dnext:
                order[i] = -1
            # Ignore the old point, it's inside
            # The new one has a chance (depending on who comes next)
            else:
                order[j] = -1
                last_idx = next_idx
                j = i
        # New angle found: this point automatically gets a chance
        # (depending on who comes next). This also means that the
        # last point (last_idx before reassignment) will make it into
        # the hull
        else:
            last_idx = next_idx
            j += 1

    # Construct the hull from all indices that are not -1
    order = order[order != -1]
    jorder = len(order) - 1
    stack = []
    j = 0
    last_idx = -1
    before_last_idx = -1
    while jorder > -1:
        next_idx = order[jorder]

        # If doing a correct turn (right), add the point to the hull
        # if doing a wrong turn (left), backtrack and skip

        # At the beginning, assume it's a good turn to start collecting points
        if j < 2:
            cp = -1
        else:
            cp = (points[last_idx, 0] - points[before_last_idx, 0]) * (
                points[next_idx, 1] - points[before_last_idx, 1]
            ) - (points[next_idx, 0] - points[before_last_idx, 0]) * (
                points[last_idx, 1] - points[before_last_idx, 1]
            )

        # turning correctly or accumulating: add to the stack
        if cp < 0:
            jorder -= 1
            stack.append(next_idx)
            j += 1
            before_last_idx = last_idx
            last_idx = next_idx

            # wrong turn: backtrack, excise wrong point and move to next vertex
        else:
            del stack[-1]
            j -= 1
            last_idx = before_last_idx
            before_last_idx = stack[j - 2] if j >= 2 else -1

    stack = np.asarray(stack)

    return stack


def _compute_group_path_with_vertex_padding(
    hull,
    points,
    transform,
    vertexpadding=10,
    points_per_curve=30,
    # TODO: check how dpi affects this
    dpi=72.0,
):
    """Offset path for a group based on vertex padding.

    At the input, the structure is [v1, v1, v1, ..., vn, vn, vn, v1]

    # NOTE: this would look better as a cubic Bezier, but ok for now.
    """
    # Short form
    ppc = points_per_curve

    # No padding, set degenerate path
    if vertexpadding == 0:
        for j, point in enumerate(hull):
            points[ppc * j : ppc * (j + 1)] = point
        points[-1] = points[0]
        return points

    # Transform into figure coordinates
    trans = transform.transform
    trans_inv = transform.inverted().transform
    points = trans(points)

    # Singleton: draw a circle around it
    if len(hull) == 1:

        # NOTE: linspace is double inclusive, which covers CLOSEPOLY
        thetas = np.linspace(
            -np.pi,
            np.pi,
            len(points),
        )
        # NOTE: dpi scaling might need to happen here
        perimeter = vertexpadding * np.vstack([np.cos(thetas), np.sin(thetas)]).T
        return trans_inv(trans(hull[0]) + perimeter)

    # Doublet: draw two semicircles
    if len(hull) == 2:

        # Unit vector connecting the two points
        dv = trans(hull[0]) - trans(hull[1])
        dv = dv / np.sqrt((dv**2).sum())

        # Draw a semicircle
        angles = np.linspace(-0.5 * np.pi, 0.5 * np.pi, 30)
        vs = np.array([np.cos(angles), -np.sin(angles), np.sin(angles), np.cos(angles)])
        vs = vs.T.reshape((len(angles), 2, 2))
        vs = np.matmul(dv, vs)

        # NOTE: dpi scaling might need to happen here
        semicircle1 = vertexpadding * vs
        semicircle2 = vertexpadding * np.matmul(vs, -np.diag((1, 1)))

        # Put it together
        vs1 = trans_inv(trans(hull[0]) + semicircle1)
        vs2 = trans_inv(trans(hull[1]) + semicircle2)
        points[:ppc] = vs1
        points[ppc:-1] = vs2
        points[-1] = points[0]
        return points

    # At least three points, i.e. a nondegenerate convex hull
    nsides = len(hull)
    for i, point1 in enumerate(hull):
        point0 = hull[i - 1]
        point2 = hull[(i + 1) % nsides]

        # NOTE: this can be optimised by computing things once
        # unit vector to previous point
        dv0 = trans(point1) - trans(point0)
        dv0 = dv0 / np.sqrt((dv0**2).sum())

        # unit vector to next point
        dv2 = trans(point2) - trans(point1)
        dv2 = dv2 / np.sqrt((dv2**2).sum())

        # span the angles
        theta0 = atan2(dv0[1], dv0[0])
        theta2 = atan2(dv2[1], dv2[0])

        # The worst that can happen is that we go exactly backwards, i.e. theta2 == theta0 + np.pi
        # if it's more than that, we are on the inside of the convex hull due to the periodicity of atan2
        if theta2 - theta0 > np.pi:
            theta2 -= 2 * np.pi

        # angles is from the point of view of the first vector, dv0
        angles = np.linspace(theta0 + np.pi / 2, theta2 + np.pi / 2, ppc)
        vs = np.array([np.cos(angles), np.sin(angles)]).T

        # NOTE: dpi scaling might need to happen here
        chunkcircle = vertexpadding * vs

        vs1 = trans_inv(trans(point1) + chunkcircle)
        points[i * ppc : (i + 1) * ppc] = vs1

    points[-1] = points[0]
    return points



"""
Animations
==========

This tutorial shows how to animate `iplotx` visualizations using `matplotlib.animation.FuncAnimation`.

For illustration purposes, we will animate a simple directed graph, rotating it around its center. We also modify the opacity of the vertices and edges, just for fun.
"""

import matplotlib.pyplot as plt
import numpy as np
import matplotlib.animation as animation
import igraph as ig
import iplotx as ipx

g = ig.Graph.Ring(3, directed=True)
layout = np.asarray(g.layout("circle").coords)

# The animation will rotate the layout of the graph
thetas = np.linspace(0, 2 * np.pi, 101)[:-1]

fig, ax = plt.subplots()
art = ipx.network(g, ax=ax, layout=layout, aspect=1)[0]
ax.set(xlim=[-2, 2], ylim=[-2, 2])


def rotate(vector, theta):
    """Rotate a 2D vector by an angle theta (in radians) clockwise."""
    return vector @ np.array(
        [
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)],
        ],
    )


def update(frame):
    # for each frame, update the vertex positions
    theta = thetas[frame]

    # The new layout is the old layout rotated by theta, clockwise
    layout_new = rotate(layout, theta)

    # Also change transparency, slower
    alpha_vertices = 1 - np.abs(np.sin(theta / 2))
    alpha_edges = 0.25 + 0.75 * np.abs(np.sin(theta / 2))

    # Edit the vertices' positions
    art.get_vertices().set_offsets(layout_new)
    art.get_vertices().set_alpha(alpha_vertices)
    art.get_edges().set_alpha(alpha_edges)
    return (art.get_vertices(), art.get_edges())


# Run the animation
ani = animation.FuncAnimation(
    fig=fig,
    func=update,
    frames=len(thetas),
    interval=30,
)



from ast import Import
import os
import unittest
import pytest
import numpy as np
import matplotlib as mpl

mpl.use("agg")
import matplotlib.pyplot as plt

import iplotx as ipx

try:
    import networkx as nx
except ImportError:
    nx = None

from utils import image_comparison


class GraphTestRunner(unittest.TestCase):
    def setUp(self):
        if nx is None:
            raise unittest.SkipTest("networkx not found, skipping tests")

    @image_comparison(baseline_images=["flat_style"], remove_text=True)
    def test_flat_style(self):
        G = nx.Graph(
            [
                (4, 1),
                (1, 2),
                (5, 3),
                (1, 3),
            ]
        )
        pos = {1: (0, 0), 2: (-1, 0.3), 3: (2, 0.17), 4: (4, 0.255), 5: (5, 0.03)}
        ipx.plot(
            G,
            layout=pos,
            vertex_labels=False,
            vertex_size=5,
            vertex_facecolor="white",
            vertex_edgecolor="black",
            vertex_linewidth=3,
            edge_linewidth=3,
            edge_color=["grey", "black"],
            margins=0.1,
        )

    @image_comparison(baseline_images=["simple_graph"], remove_text=True)
    def test_simple_graph(self):
        G = nx.Graph()
        G.add_edge(1, 2)
        G.add_edge(1, 3)
        G.add_edge(1, 5)
        G.add_edge(2, 3)
        G.add_edge(3, 4)
        G.add_edge(4, 5)

        # explicitly set positions
        pos = {1: (0, 0), 2: (-1, 0.3), 3: (2, 0.17), 4: (4, 0.255), 5: (5, 0.03)}

        ipx.plot(
            G,
            layout=pos,
            vertex_labels=True,
            style={
                "vertex": {
                    "size": "label",
                    "facecolor": "white",
                    "edgecolor": "black",
                    "linewidth": 5,
                    "label": {
                        "size": 36,
                        "color": "black",
                    },
                },
                "edge": {
                    "linewidth": 5,
                },
            },
            margins=0.2,
        )

    @image_comparison(baseline_images=["cluster-layout"], remove_text=True)
    def test_cluster_layout(self):
        G = nx.davis_southern_women_graph()
        communities = nx.community.greedy_modularity_communities(G)

        # Layout does not appear deterministic, so we use a fixed layout from one run of spring_layout
        pos = {
            "Evelyn Jefferson": np.array([-25.68131066, -10.80854424]),
            "Laura Mandeville": np.array([-25.55280383, -10.99674201]),
            "Theresa Anderson": np.array([-25.38187247, -10.86875164]),
            "Brenda Rogers": np.array([-25.67901346, -11.17009239]),
            "Charlotte McDowd": np.array([-25.85576192, -11.09635554]),
            "Frances Anderson": np.array([-24.97100665, -10.80341997]),
            "Eleanor Nye": np.array([-25.22178561, -11.61473984]),
            "Pearl Oglethorpe": np.array([-24.43628739, -11.25006292]),
            "Ruth DeSand": np.array([-25.15041064, -11.82194973]),
            "E1": np.array([-26.10000875, -11.13170858]),
            "E2": np.array([-25.52047415, -10.41822091]),
            "E3": np.array([-25.48205106, -10.7056092]),
            "E4": np.array([-25.96271954, -10.75283759]),
            "E5": np.array([-25.35150057, -11.23805354]),
            "E6": np.array([-25.1129681, -11.11829674]),
            "E7": np.array([-25.5206234, -11.4515497]),
            "E13": np.array([-29.38952624, -9.82903937]),
            "Nora Fayette": np.array([-30.14866675, -9.89241409]),
            "Olivia Carleton": np.array([-30.44992786, -8.93390847]),
            "Katherina Rogers": np.array([-29.87511453, -10.05207037]),
            "Helen Lloyd": np.array([-30.70287724, -10.18706022]),
            "E12": np.array([-29.97554058, -10.44474811]),
            "E14": np.array([-30.32816285, -10.28626239]),
            "Sylvia Avondale": np.array([-29.68181361, -10.1960916]),
            "Myra Liddel": np.array([-29.50794269, -10.44247286]),
            "E9": np.array([-29.86346377, -9.54678644]),
            "Flora Price": np.array([-30.07123751, -8.8956997]),
            "E10": np.array([-30.15355559, -10.54517655]),
            "E11": np.array([-30.59292037, -9.39236595]),
            "E8": np.array([-19.26284314, -9.57050676]),
            "Dorothy Murchison": np.array([-19.48148306, -10.57263599]),
            "Verne Sanderson": np.array([-19.0455969, -8.57476521]),
        }

        # Nodes colored by cluster
        node_color = {}
        for nodes, clr in zip(communities, ("tab:blue", "tab:orange", "tab:green")):
            for node in nodes:
                node_color[node] = clr

        fig, ax = plt.subplots(figsize=(6, 6))
        ipx.plot(
            G,
            layout=pos,
            style={
                "vertex": {
                    "facecolor": node_color,
                    "linewidth": 0,
                    "size": 21,
                }
            },
            ax=ax,
        )

    @image_comparison(
        baseline_images=["directed_graph_with_colorbar"], remove_text=True
    )
    def test_directed_graph(self):
        seed = 13648  # Seed random number generators for reproducibility
        G = nx.random_k_out_graph(10, 3, 0.5, seed=seed)
        pos = nx.spring_layout(G, seed=seed)

        node_sizes = [3 + 1.5 * i for i in range(len(G))]
        M = G.number_of_edges()
        edge_colors = list(range(2, M + 2))
        edge_alphas = [(5 + i) / (M + 4) for i in range(M)]
        cmap = plt.cm.plasma

        fig, ax = plt.subplots(figsize=(6, 4))

        arts = ipx.plot(
            network=G,
            ax=ax,
            layout=pos,
            style={
                "vertex": {
                    "size": node_sizes,
                    "facecolor": "indigo",
                    "edgecolor": "none",
                },
                "edge": {
                    "color": edge_colors,
                    "alpha": edge_alphas,
                    "cmap": cmap,
                    "linewidth": 2,
                    "paralleloffset": 0,
                    "arrow": {
                        "marker": ">",
                        "width": 5,
                    },
                },
            },
        )
        fig.colorbar(
            arts[0].get_edges(),
            ax=ax,
        )

    @image_comparison(baseline_images=["empty_graph"], remove_text=True)
    def test_display_empty_graph(self):
        G = nx.empty_graph()
        fig, ax = plt.subplots()
        ipx.plot(G, ax=ax)

    @image_comparison(baseline_images=["shortest_path"], remove_text=True)
    def test_display_shortest_path(self):
        G = nx.Graph()
        G.add_nodes_from(["A", "B", "C", "D", "E", "F", "G", "H"])
        G.add_edge("A", "B", weight=4)
        G.add_edge("A", "H", weight=8)
        G.add_edge("B", "C", weight=8)
        G.add_edge("B", "H", weight=11)
        G.add_edge("C", "D", weight=7)
        G.add_edge("C", "F", weight=4)
        G.add_edge("C", "I", weight=2)
        G.add_edge("D", "E", weight=9)
        G.add_edge("D", "F", weight=14)
        G.add_edge("E", "F", weight=10)
        G.add_edge("F", "G", weight=2)
        G.add_edge("G", "H", weight=1)
        G.add_edge("G", "I", weight=6)
        G.add_edge("H", "I", weight=7)

        # Find the shortest path from node A to node E
        path = nx.shortest_path(G, "A", "E", weight="weight")

        # Create a list of edges in the shortest path
        path_edges = list(zip(path, path[1:]))
        nx.set_node_attributes(G, nx.spring_layout(G, seed=37), "pos")
        nx.set_edge_attributes(
            G,
            {
                (u, v): {
                    "color": (
                        "red"
                        if (u, v) in path_edges or tuple(reversed((u, v))) in path_edges
                        else "black"
                    ),
                    "label": d["weight"],
                }
                for u, v, d in G.edges(data=True)
            },
        )

        fig, ax = plt.subplots()
        ipx.plot(
            G,
            ax=ax,
            layout="pos",
            vertex_labels=True,
            edge_labels=True,
        )

    @image_comparison(baseline_images=["house_with_colors"], remove_text=True)
    def test_display_house_with_colors(self):
        G = nx.house_graph()
        fig, ax = plt.subplots(figsize=(4, 4))
        nx.set_node_attributes(
            G, {0: (0, 0), 1: (1, 0), 2: (0, 1), 3: (1, 1), 4: (0.5, 2.0)}, "pos"
        )
        nx.set_node_attributes(
            G,
            {
                n: {
                    "size": 40 if n != 4 else 30,
                    "color": "tab:blue" if n != 4 else "tab:orange",
                }
                for n in G.nodes()
            },
        )
        ipx.plot(
            G,
            ax=ax,
            layout="pos",
            style={
                "edge": {
                    "alpha": 0.5,
                    "linewidth": 6,
                },
                "vertex": {
                    "size": G.nodes.data("size"),
                    "facecolor": G.nodes.data("color"),
                    "edgecolor": "k",
                },
            },
            margins=0.1,
        )
        plt.tight_layout()

    @image_comparison(baseline_images=["labels_and_colors"], remove_text=True)
    def test_labels_and_colors(self):
        """Test complex labels and colors."""
        G = nx.cubical_graph()
        pos = nx.spring_layout(G, seed=3113794652)  # positions for all nodes
        nx.set_node_attributes(G, pos, "pos")  # Will not be needed after PR 7571
        labels = iter(
            [
                r"$a$",
                r"$b$",
                r"$c$",
                r"$d$",
                r"$\alpha$",
                r"$\beta$",
                r"$\gamma$",
                r"$\delta$",
            ]
        )
        nx.set_node_attributes(
            G,
            {
                n: {
                    "color": "tab:red" if n < 4 else "tab:blue",
                    "label": next(labels),
                }
                for n in G.nodes()
            },
        )

        fig, ax = plt.subplots(figsize=(5, 4))
        ipx.plot(
            G,
            ax=ax,
            layout="pos",
            vertex_labels=True,
            style={
                "vertex": {
                    "facecolor": G.nodes.data("color"),
                    "alpha": G.nodes.data("alpha"),
                    "size": 30,
                    "label": {"size": 22, "color": "whitesmoke"},
                    "zorder": 7,
                },
                "edge": {
                    "color": "tab:grey",
                },
            },
        )
        for i in range(2):
            subG = G.subgraph(range(i * 4, (i + 1) * 4))
            ipx.plot(
                subG,
                layout="pos",
                ax=ax,
                style={
                    "edge": {
                        "color": ["tab:red", "tab:blue"][i],
                        "alpha": 0.5,
                        "linewidth": 8,
                        "zorder": 0,
                    },
                    "vertex": {
                        "size": 25,
                        "alpha": 0,
                    },
                },
                margins=0.1,
            )
        plt.tight_layout()

    @image_comparison(baseline_images=["complex"], remove_text=True)
    def test_complex(self):
        import itertools as it

        nodes = "ABC"
        prod = list(it.product(nodes, repeat=2)) * 4
        G = nx.MultiDiGraph()
        for i, (u, v) in enumerate(prod):
            G.add_edge(u, v, w=round(i / 3, 2))
        nx.set_node_attributes(G, nx.spring_layout(G, seed=3113794652), "pos")
        csi = it.cycle([5 * r for r in it.accumulate([0.15] * 4)])
        nx.set_edge_attributes(G, {e: next(csi) for e in G.edges(keys=True)}, "tension")
        nx.set_edge_attributes(
            G,
            {tuple(e): w for *e, w in G.edges(keys=True, data="w")},
            "label",
        )

        fig, ax = plt.subplots()
        ipx.plot(
            G,
            ax=ax,
            layout="pos",
            edge_labels=True,
            margins=0.03,
            style={
                "edge": {
                    "curved": True,
                    "tension": G.edges.data("tension"),
                    "color": G.edges.data("w"),
                    "cmap": mpl.colormaps["inferno"],
                    "linewidth": 1,
                    "looptension": 7.5,
                    "label": {
                        "color": "black",
                        "bbox": {
                            "facecolor": "none",
                        },
                    },
                },
            },
        )


def suite():
    return unittest.TestSuite(
        [
            unittest.defaultTestLoader.loadTestsFromTestCase(GraphTestRunner),
        ]
    )


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""
Minimum spanning trees
======================

This example from igraph shows how to work with edge labels and linewidths.
"""

import random
import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx


random.seed(0)
g = ig.Graph.Lattice([5, 5], circular=False)
g.es["weight"] = [random.randint(1, 20) for _ in g.es]

mst_edges = g.spanning_tree(weights=g.es["weight"], return_tree=False)

g.es["color"] = "lightgray"
g.es[mst_edges]["color"] = "midnightblue"
g.es["width"] = 1.0
g.es[mst_edges]["width"] = 3.0

layout = g.layout("grid")

ipx.plot(
    g,
    layout=layout,
    edge_labels=g.es["weight"],
    style={
        "vertex": {
            "facecolor": "lightblue",
            "edgecolor": "black",
            "linewidth": 1,
        },
        "edge": {
            "linewidth": g.es["width"],
            "label": {
                "rotate": True,
                "bbox": {
                    "facecolor": "white",
                    "edgecolor": "none",
                },
            },
        },
    },
)

# %%
# If you want to avoid rotating the labels vertically and let them follow the edge directions instead,
# just set the "rotate" parameter to False (or skip it in the default style):

ipx.plot(
    g,
    layout=layout,
    edge_labels=g.es["weight"],
    style={
        "vertex": {
            "facecolor": "lightblue",
            "edgecolor": "black",
            "linewidth": 1,
        },
        "edge": {
            "linewidth": g.es["width"],
            "label": {
                "rotate": False,
                "bbox": {
                    "facecolor": "white",
                    "edgecolor": "none",
                },
            },
        },
    },
)



"""
Module for edge arrows in iplotx.
"""

import numpy as np
import matplotlib as mpl
from matplotlib.patches import PathPatch

from ..style import (
    get_style,
    rotate_style,
)


class EdgeArrowCollection(mpl.collections.PatchCollection):
    """Collection of arrow patches for plotting directed edgs."""

    _factor = 1.0

    def __init__(
        self,
        edge_collection,
        *args,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        **kwargs,
    ) -> None:
        """Initialize the edge arrow collection.

        Parameters:
            edge_collection: The edge collection to which these arrows belong.
            transform: The transform to apply to the arrows. This related to the arrow size
                scaling, not the arrow tip position which is controlled by set_offset_transform.
        """

        self._edge_collection = edge_collection
        self._style = get_style(".edge.arrow")

        patches, sizes = self._create_artists()

        if "cmap" in self._edge_collection._style:
            kwargs["cmap"] = self._edge_collection._style["cmap"]
            kwargs["norm"] = self._edge_collection._style["norm"]

        super().__init__(
            patches,
            offsets=np.zeros((len(patches), 2)),
            offset_transform=self.get_offset_transform(),
            transform=transform,
            match_original=True,
            *args,
            **kwargs,
        )
        self._angles = np.zeros(len(self._paths))

        # Compute _transforms like in _CollectionWithScales for dpi issues
        self.set_sizes(sizes)

    def get_sizes(self):
        """Get arrow sizes (max of width and height), not scaled by dpi."""
        return self._sizes

    def get_sizes_dpi(self):
        """Get arrow sizes (max of width and height) scaled by dpi."""
        return self._transforms[:, 0, 0]

    def set_sizes(self, sizes, dpi=72.0):
        """Set vertex sizes.

        This rescales the current vertex symbol/path linearly, using this
        value as the largest of width and height.

        @param sizes: A sequence of vertex sizes or a single size.
        """
        if sizes is None:
            self._sizes = np.array([])
            self._transforms = np.empty((0, 3, 3))
        else:
            self._sizes = np.asarray(sizes)
            self._transforms = np.zeros((len(self._sizes), 3, 3))
            scale = self._sizes * dpi / 72.0 * self._factor
            self._transforms[:, 0, 0] = scale
            self._transforms[:, 1, 1] = scale
            self._transforms[:, 2, 2] = 1.0
        self.stale = True

    get_size = get_sizes
    set_size = set_sizes

    def set_figure(self, fig) -> None:
        """Set the figure for this artist and all children."""
        super().set_figure(fig)
        self.set_sizes(self._sizes, self.get_figure(root=True).dpi)
        for child in self.get_children():
            child.set_figure(fig)

    def get_offset_transform(self):
        """Get offset transform for the edge arrows. This sets the tip of each arrow."""
        return self._edge_collection.get_transform()

    get_size = get_sizes
    set_size = set_sizes

    def _create_artists(self):
        style = self._style if self._style is not None else {}

        patches = []
        sizes = []
        for i in range(len(self._edge_collection._vertex_ids)):
            stylei = rotate_style(style, index=i)
            if ("facecolor" not in stylei) and ("color" not in stylei):
                stylei["facecolor"] = self._edge_collection.get_edgecolors()[i][:3]
            if ("edgecolor" not in stylei) and ("color" not in stylei):
                stylei["edgecolor"] = self._edge_collection.get_edgecolors()[i][:3]
            if "alpha" not in stylei:
                stylei["alpha"] = self._edge_collection.get_edgecolors()[i][3]
            if "linewidth" not in stylei:
                stylei["linewidth"] = self._edge_collection.get_linewidths()[i]

            patch, size = make_arrow_patch(
                **stylei,
            )
            patches.append(patch)
            sizes.append(size)

        return patches, sizes

    def set_array(self, A):
        """Set the array for cmap/norm coloring, but keep the facecolors as set (usually 'none')."""
        raise ValueError("Setting an array for arrows directly is not supported.")

    def set_colors(self, colors):
        """Set arrow colors (edge and/or face) based on a colormap."""
        # NOTE: facecolors is always an array because we come from patches
        # It can have zero alpha (i.e. if we choose "none", or a hollow marker)
        self.set_edgecolor(colors)
        has_facecolor = self._facecolors[:, 3] > 0
        self._facecolors[has_facecolor] = colors[has_facecolor]

    @mpl.artist.allow_rasterization
    def draw(self, renderer):
        self.set_sizes(self._sizes, self.get_figure(root=True).dpi)
        super().draw(renderer)


def make_arrow_patch(marker: str = "|>", width: float = 8, **kwargs):
    """Make a patch of the given marker shape and size."""
    height = kwargs.pop("height", width * 1.3)

    # Normalise by the max size, this is taken care of in _transforms
    # subsequently in a way that is nice to dpi scaling
    size_max = max(width, height)
    if size_max > 0:
        height /= size_max
        width /= size_max

    if marker == "|>":
        codes = ["MOVETO", "LINETO", "LINETO", "CLOSEPOLY"]
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        path = mpl.path.Path(
            np.array(
                [
                    [-height, width * 0.5],
                    [-height, -width * 0.5],
                    [0, 0],
                    [-height, width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == ">":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO"]
        path = mpl.path.Path(
            np.array([[-height, width * 0.5], [0, 0], [-height, -width * 0.5]]),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == ">>":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        overhang = kwargs.pop("overhang", 0.25)
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [0, 0],
                    [-height, -width * 0.5],
                    [-height * (1.0 - overhang), 0],
                    [-height, width * 0.5],
                    [0, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == ")>":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        overhang = kwargs.pop("overhang", 0.25)
        codes = ["MOVETO", "LINETO", "CURVE3", "CURVE3", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [0, 0],
                    [-height, -width * 0.5],
                    [-height * (1.0 - overhang), 0],
                    [-height, width * 0.5],
                    [0, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == ")":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "CURVE3", "CURVE3"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height * 0.5, width * 0.5],
                    [height * 0.5, 0],
                    [-height * 0.5, -width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == "|":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO"]
        path = mpl.path.Path(
            np.array([[-height, width * 0.5], [-height, -width * 0.5]]),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == "s":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height, width * 0.5],
                    [-height, -width * 0.5],
                    [0, -width * 0.5],
                    [0, width * 0.5],
                    [-height, width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "d":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height * 0.5, width * 0.5],
                    [-height, 0],
                    [-height * 0.5, -width * 0.5],
                    [0, 0],
                    [-height * 0.5, width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "p":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height, 0],
                    [0, 0],
                    [0, -width],
                    [-height, -width],
                    [-height, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "q":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height, 0],
                    [0, 0],
                    [0, width],
                    [-height, width],
                    [-height, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "none":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO"]
        path = mpl.path.Path(
            np.array(
                [
                    [0, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    else:
        raise ValueError(f"Arrow marker not found: {marker}.")

    patch = PathPatch(
        path,
        **kwargs,
    )
    return patch, size_max



"""
Company structure
=================

This example shows how to use `iplotx` to visualise simple logical relationships, for instance the organisational chart of a company.
"""

import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

g = nx.DiGraph()
g.add_edges_from(
    [
        ("CEO", "Manager 1"),
        ("CEO", "Manager 2"),
        ("CEO", "Manager 3"),
        ("Manager 1", "Employee 1"),
        ("Manager 1", "Employee 2"),
        ("Manager 2", "Employee 3"),
        ("Manager 2", "Employee 4"),
        ("Manager 3", "Employee 4"),
    ]
)
layout = {
    "CEO": [0, 0],
    "Manager 1": [1, -1],
    "Manager 2": [1, 0],
    "Manager 3": [1, 1],
    "Employee 1": [2, -1],
    "Employee 2": [2, -0.5],
    "Employee 3": [2, 0],
    "Employee 4": [2, 0.5],
}
fig, ax = plt.subplots()
ipx.network(
    g,
    layout=layout,
    ax=ax,
    vertex_labels=True,
    style="hollow",
    vertex_marker="r",
    edge_ports=[("e", "w")] * 8,
    edge_curved=True,
    edge_tension=[0.65] * 3 + [0.6] * 5,
    margins=(0.08, 0.05),
)
# Show the layout top to bottom for clarity
ax.invert_yaxis()

# %%
# If you prefer a more squared look, you can use S-style waypoints with per-edge offsets:

fig, ax = plt.subplots()
ipx.network(
    g,
    layout=layout,
    ax=ax,
    vertex_labels=True,
    style="hollow",
    vertex_marker="r",
    edge_ports=[("e", "w")] * 8,
    edge_waypoints="xmidy0,xmidy1",
    edge_offset=[(0, 0)] * 7 + [(0, -6)],
    margins=(0.08, 0.05),
    edge_linestyle=["-"] * 7 + ["--"],
)
ax.invert_yaxis()

# %%
# The `offset` edge style bypasses any layout coordinate system and acts in figure points directly.
# Combining edge offsets with noncartesian layout coordinate systems (e.g. `polar`) can lead to unexpected results.



"""
This module focuses on how to ingest network data into standard data structures no matter what library they come from.
"""

import pathlib
import pkgutil
import importlib
import warnings
from typing import (
    Optional,
    Sequence,
    Protocol,
)
from collections.abc import Hashable
import pandas as pd

from ..typing import (
    GraphType,
    LayoutType,
    TreeType,
)
from .typing import (
    NetworkDataProvider,
    NetworkData,
    TreeDataProvider,
    TreeData,
)

provider_protocols = {
    "network": NetworkDataProvider,
    "tree": TreeDataProvider,
}

# Internally supported data providers
data_providers: dict[str, dict[str, Protocol]] = {
    kind: {} for kind in provider_protocols
}
for kind in data_providers:
    providers_path = pathlib.Path(__file__).parent.joinpath("providers").joinpath(kind)
    for importer, module_name, _ in pkgutil.iter_modules([providers_path]):
        module = importlib.import_module(
            f"iplotx.ingest.providers.{kind}.{module_name}"
        )
        for key, val in module.__dict__.items():
            if key == provider_protocols[kind].__name__:
                continue
            if key.endswith("DataProvider"):
                data_providers[kind][module_name] = val
                break
    del providers_path


def network_library(network) -> str:
    """Guess the network library used to create the network."""
    for name, provider in data_providers["network"].items():
        if provider.check_dependencies():
            graph_type = provider.graph_type()
            if isinstance(network, graph_type):
                return name
    raise ValueError(
        f"Network {network} did not match any available network library.",
    )


def tree_library(tree) -> str:
    """Guess the tree library used to create the tree."""
    for name, provider in data_providers["tree"].items():
        if provider.check_dependencies():
            tree_type = provider.tree_type()
            if isinstance(tree, tree_type):
                return name
    raise ValueError(
        f"Tree {tree} did not match any available tree library.",
    )


# Functions to ingest data from various libraries
def ingest_network_data(
    network: GraphType,
    layout: Optional[LayoutType] = None,
    vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
    edge_labels: Optional[Sequence[str] | dict[str,]] = None,
) -> NetworkData:
    """Create internal data for the network."""
    _update_data_providers("network")

    nl = network_library(network)

    if nl in data_providers["network"]:
        provider: NetworkDataProvider = data_providers["network"][nl]
    else:
        sup = ", ".join(data_providers["network"].keys())
        raise ValueError(
            f"Network library '{nl}' is not installed. "
            f"Currently installed supported libraries: {sup}."
        )

    result = provider()(
        network=network,
        layout=layout,
        vertex_labels=vertex_labels,
        edge_labels=edge_labels,
    )
    result["network_library"] = nl
    return result


def ingest_tree_data(
    tree: TreeType,
    layout: Optional[str] = "horizontal",
    orientation: Optional[str] = None,
    directed: bool | str = False,
    layout_style: Optional[dict[str, str | int | float]] = None,
    vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
    edge_labels: Optional[Sequence[str] | dict[Hashable, str]] = None,
    leaf_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
) -> TreeData:
    """Create internal data for the tree."""
    _update_data_providers("tree")

    tl = tree_library(tree)

    if tl in data_providers["tree"]:
        provider: TreeDataProvider = data_providers["tree"][tl]
    else:
        sup = ", ".join(data_providers["tree"].keys())
        raise ValueError(
            f"Tree library '{tl}' is not installed. "
            f"Currently installed supported libraries: {sup}."
        )

    result = provider(
        tree=tree,
    )(
        layout=layout,
        orientation=orientation,
        directed=directed,
        layout_style=layout_style,
        vertex_labels=vertex_labels,
        edge_labels=edge_labels,
        leaf_labels=leaf_labels,
    )
    result["tree_library"] = tl

    # TODO: cascading thing here

    return result


# INTERNAL FUNCTIONS
def _update_data_providers(kind):
    """Update data provieders dynamically from external packages."""
    discovered_providers = importlib.metadata.entry_points(
        group=f"iplotx.{kind}_data_providers"
    )
    for entry_point in discovered_providers:
        if entry_point.name not in data_providers["network"]:
            try:
                data_providers[kind][entry_point.name] = entry_point.load()
            except Exception as e:
                warnings.warn(
                    f"Failed to load {kind} data provider '{entry_point.name}': {e}"
                )
